import java.net.*;
import java.rmi.*;
import java.rmi.server.*;
import java.rmi.registry.*;
import java.util.*;

public class Server extends UnicastRemoteObject implements Server_itf {
	
	//Access ServerObjects from ID
	private HashMap<Integer,ServerObject> idObjects;
	
	//Access ServerObjects from Name
	private HashMap<String,Integer> string2id;
	
	//Acces to class of the object from ID
	private HashMap<Integer,SharedInfo> idClass;
	
	//Number of created objects
	private int count;
	
		
	public Server() throws RemoteException {
		super();
		this.idObjects = new HashMap<Integer,ServerObject>();
		this.string2id = new HashMap<String,Integer>();
		this.idClass = new HashMap<Integer,SharedInfo>();
		this.count = 0;
	}
	
	//Get the indice of the object from its name
	public SharedInfo lookup(String name) {
		if (string2id.containsKey(name)) {
			int id = string2id.get(name);
			return idClass.get(id);
		} else {
			return new SharedInfo(-1,"");
		}
	}
	
	//SAFE METHOD
	//To be used if many SharedObject are created simultaneously
	//Get the indice of the object from its name
	// /!\ A "lookup" must be follwed by a creation. /!\
	public SharedInfo lookupSafe(String name) {
		
		int result = -1;
		SharedInfo ret;
		
		synchronized(this){
			if (string2id.containsKey(name)) {
				result = string2id.get(name);				
				if (result == -1){
					//The object is being created (but not yet)
					//We will wait for the creation...
					try {
						while(string2id.get(name) == -1) {
							System.out.println("Waiting for the creation...");
							this.wait();
						}
					// finally, object has been initialized
					result = string2id.get(name);					
					} catch (Exception e) {
						e.printStackTrace();
					}
				}
				ret = this.idClass.get(result);
				
			} else {
				//for synchronization purpose : we store
				//the name, so no matters how long the
				//client will take to create the object, it
				//will be the only one to be able to create it.
				string2id.put(name, -1);
				ret = new SharedInfo(-1,name);
			}
		}			
		return ret;
	}
	
	//association of the ID to the name
	public void register(String name, int id) {
		System.out.println("Registering object "+name+",#"+id);
		this.string2id.put(name,id);
		//We wake up all the threads which were waiting for
		//the creation of the object. They will fall asleep
		//by themselves if it wasn't the object they look after.
		synchronized(this){
			this.notifyAll();
		}

	}
	
	//we create a new ServerObject with a unique  ID
	public int create(Object o) {
		int id;
		synchronized (this) {
			count += 1;
			id = count;
		}
		ServerObject newObj = new ServerObject(id,o);
		this.idObjects.put(id,newObj);
		String className = o.getClass().getName();
		this.idClass.put(id,new SharedInfo(id,className));
		
		return id;
	}
	
	public Object lock_read(int id, Client_itf client) {
		System.out.println("Appel à lock_read sur " + id);
		ServerObject so = idObjects.get(id);
		return so.lock_read(client);
		
	}
	
	public Object lock_write(int id, Client_itf client) {
		System.out.println("Appel à lock_write sur " + id);
		ServerObject so = idObjects.get(id);
		return so.lock_write(client);
	}
	
	
	 public static void main(String args[]) {
	    try {
	    	Registry registry = LocateRegistry.createRegistry(1099);
			Server server = new Server();
			String hostname = InetAddress.getLocalHost().getHostName();
			Naming.rebind("//"+hostname+"/Server", server);
			System.out.println("Server listening on "+hostname+"...");
		} catch (Exception ex) {
	    		ex.printStackTrace();
	    	}
	    }

}
